# -*- coding: utf-8 -*-

"""
Daemon objects
==============

This module defines daemons classes, those objects represent the background
processes, and specifies how to launch and stop them and query their state.

Daemons can be ``start``ed to double fork and run in the background. They can
``run`` in the foreground. Started daemons can be ``stop``ped and
``restart``ed.
"""

import os
import signal
import time
import subprocess
import collections

from paver.easy import path, info, debug, error
from paver.deps.six import string_types
from paver.shell import _shlex_quote

from sett import ROOT
from sett.daemon_tasks import daemon_task, ctl_task  # noqa


def shlex_quote(args):
    return ' '.join(_shlex_quote(quoted) for quoted in args)


class DaemonGroup(object):
    """
    A group of Daemon
    """
    def __init__(self, daemons, name=None):
        self._daemons = daemons
        self.name = name

    def __eq__(self, other):
        return (isinstance(other, DaemonGroup) and
                set(self._daemons) == set(other._daemons))

    def __repr__(self):
        if self.name:
            return '<Daemons "{}">'.format(self.name)
        return '<Daemons {}>'.format(', '.join(d.name for d in self))

    def __iter__(self):
        return iter(self._daemons)

    def __len__(self):
        return len(self._daemons)

    def run(self, *args):
        """
        Run the daemon in the foreground. It must be a Group of only 1 daemon.
        """
        assert len(self) == 1
        self._daemons[0].run(*args)

    @property
    def reloadable(self):
        return any(d.reloadable for d in self)

    def reload(self):
        """
        Reload all the reloadable daemons
        """
        for d in self:
            if d.reloadable():
                d.reload()

    def start(self):
        """
        Start all the daemons of the group
        """
        for d in self:
            d.start()

    def stop(self):
        """
        Stop all the daemons of the group
        """
        for d in self:
            d.stop()

    def status(self):
        """
        Show the statuses of the daemons of the group
        """
        return [d.status() for d in self]


class Daemons(object):
    """
    A set of ``Daemon``s, some of which can be grouped.

    ``Daemons`` can register a set of daemons with unique names, and a set of
    groups of daemons. Groups and daemons name must be unique.

    The instance is subscribable and return ``DaemonGroup`` instances (even for
    single daemon, it returns a ``DaemonGroup`` of length 1).
    """
    def __init__(self, *daemons, **groups):
        self._daemons = {}
        self._daemons_groups = collections.defaultdict(list)
        self._names = set([None])

        self.register_all(daemons)
        for group, daemons in groups.items():
            self.register_all(daemons, group=group)

    def __repr__(self):
        if not self:
            return '<No daemons>'
        return '<Daemons {}>'.format(','.join(d.name for d in self))

    def __bool__(self):
        return bool(self._daemons)

    __nonzero__ = __bool__

    def __iter__(self):
        return iter(self._daemons_groups[None])

    def groups(self):
        """
        Return the groups of daemons
        """
        return (self[g] for g in self._daemons_groups if g)

    def __contains__(self, value):
        return value in self._daemons or value in self._daemons_groups

    def __getitem__(self, name):
        if name == 'all':
            name = None
        if name not in self._names:
            raise KeyError(name)
        if name in self._daemons_groups:
            return DaemonGroup(self._daemons_groups[name], name=name)
        return DaemonGroup([self._daemons[name]])

    def register_all(self, daemons, group=None):
        """
        Register a set of daemons. If group is not `None`, it is the name of
        the group containing all the daemons.
        """
        for d in daemons:
            self.register(d, group)

    def register(self, daemon, group=None):
        """
        Register a daemon, if group is not `None` the daemon is also registed
        in the given group. If the groupd does not exist it is created.
        """
        assert group is None or isinstance(group, str) and group != 'all'

        if daemon.name in self._names:
            raise ValueError('Daemon name {} is already registered'.format(daemon.name))
        if group in self._daemons:
            raise ValueError('Group name {} is already registered by a daemon'.format(group))
        self._names.add(daemon.name)

        self._daemons_groups[None].append(daemon)
        self._daemons[daemon.name] = daemon
        if group:
            self._names.add(group)
            self._daemons_groups[group].append(daemon)


class DaemonError(Exception):
    pass


class SignalSender(object):
    if hasattr(signal, 'Signal'):
        @classmethod
        def is_signal(self, value):
            #  Python 3
            return value in signal.Signal
    else:
        @classmethod
        def is_signal(self, value):
            #  Python 2
            try:
                signal.getsignal(value)
            except ValueError:
                return False
            else:
                return True

    def __init__(self, signal):
        self.signal = signal

    def __call__(self, pid):
        try:
            os.kill(pid, self.signal)
        except OSError as e:
            raise DaemonError(e)


class Daemon(object):
    """
    A single background process.

    *cmd* is the list of args of the command that **run** the daemon.
    *daemonize* is a callable or a list. It defines how the process goes in the
    background. When *daemonize* is a list, it is appended to the *cmd* before
    running. If it's a callable, it's called with the *pid_file* and the result
    should be an iterable that is appended to the *cmd*. If it's *None*, then
    the *cmd* daemonize itself by default and **run** is not enabled (it may
    only be started in the background).

    *pid_file* defines which file will store the PID. If it is `None`, the pid
    file is automatically determined from the name of the daemon.

    *reload* may be a signal descriptor (ex: 'SIGHUP', 1, signal.SIGHUP) or
    None. If it's None, reloading the process is disabled, else this signal is
    send when a reload is ordered.

    *env* is a `dict` of value that are added to the environ before running.

    *name* is the name of the daemon. If it's not given, it's guessed from the cmd.
    """
    def __init__(self, cmd, daemonize=[], pid_file=None, env=None, name=None, reload=None, kill=None, description=''):
        self.name = name or str(path(cmd[0]).basename())
        self.description = description or self.name
        self.cmd = cmd
        self.daemonize = daemonize
        self.env = env
        self._pid_file = pid_file

        self.kill_handler = self.make_signal_handler(kill) or self._default_killer
        self.reload_handler = self.make_signal_handler(reload)

    @classmethod
    def make_signal_handler(self, signal_value):
        if callable(signal_value):
            return signal_value

        if isinstance(signal_value, string_types) and signal_value.startswith('SIG'):
            signal_value = getattr(signal, signal_value)

        if signal_value is None:
            return None
        if SignalSender.is_signal(signal_value):
            return SignalSender(signal_value)

        raise TypeError('Unexpected {}({!r}), expected a signal or a signal name'.format(type(signal_value), signal_value))

    @property
    def environ(self):
        """
        The update to the environ of the process
        """
        return self.env or {}

    @property
    def command(self):
        """
        The command as a string.
        """
        return shlex_quote(self.get_command())

    @property
    def daemon_command(self):
        """
        The daemonize command as a string.
        """
        return shlex_quote(self.get_daemon_command())

    @property
    def pid_file(self):
        """
        The pid file
        """
        return self._pid_file or ROOT.joinpath('var/run/', self.name + '.pid')

    @property
    def reloadable(self):
        """
        Indicates if the daemon handle reloading
        """
        return bool(self.reload_handler)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<Daemon {}>'.format(self.name)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return (isinstance(other, Daemon) and
                self.name == other.name and
                self.cmd == other.cmd and
                self.daemonize == other.daemonize and
                self.env == other.env)

    def reload(self):
        """
        Reloads the daemon if it supports it.

        If the daemon cannot be reload or if the reload fails it raises a
        ``DaemonError``.
        """
        if not self.reloadable:
            raise DaemonError('Daemon %s is not reloadable' % self)

        pid = self._get_pid()
        if not pid:
            raise DaemonError('Daemon %s is not running' % self)

        self.reload_handler(pid)

    def status(self):
        """
        Query the state of the daemon.
        """
        pid = self._get_pid()
        if not pid:
            return 'Not running'
        try:
            os.kill(pid, 0)
        except OSError:
            return 'Not running, stale PID file'
        else:
            return 'Running with pid {}'.format(pid)

    def start(self):
        """
        Starts in the background
        """
        info('Starting %s', self)
        pid_dir = path(self.pid_file).dirname()
        if not pid_dir.exists():
            pid_dir.makedirs()

        process = self._run(self.get_daemon_command())
        if self.daemonize:
            debug('Waiting for process')
            process.wait()
        elif self.daemonize is None:
            debug('Writing pid in %s', self.pid_file)
            self._set_pid(process.pid)

    def get_daemon_command(self):
        """
        Returns a list of the args to run the process daemonized
        """
        daemonize = self.daemonize
        command = list(self.cmd)

        if callable(daemonize):
            daemonize = daemonize(self.pid_file)
            command.extend(daemonize)
        elif isinstance(daemonize, (list, tuple)):
            command.extend(daemonize)
        return command

    def get_command(self):
        """
        Returns a list of the args to run the process in foreground
        """
        return self.cmd

    def run(self):
        """
        Runs the daemon in foreground
        """
        return self._run(self.get_command()).wait()

    def _run(self, command):
        if self.env:
            env = dict(os.environ)
            env.update(self.env)
        else:
            env = os.environ

        info('Running %s', ' '.join(command))
        return subprocess.Popen(command, env=env)

    def stop(self):
        """
        Stops the daemon by sending it a SIGTERM, then a SIGKILL if it has not
        terminated quickly.

        If the daemon is not running, fails silently, else it raises a ``DaemonError``
        """
        info('Stopping %s', self)
        pid = self._get_pid()
        if pid is None:
            info('%s had no PID file or was not running', self.name)
            return
        self.kill_handler(pid)

    def _default_killer(self, pid):
        try:
            os.kill(pid, signal.SIGKILL)
        except OSError:
            info('%s was not running', self.name)
            self._set_pid(None)
            return

        for x in range(5, 11):
            time.sleep(2 ** x / 500)
            try:
                os.kill(pid, 0)
            except OSError:
                break
        else:
            info('Program %s did not respond to SIGKILL, sending SIGTERM', pid)
            os.kill(pid, signal.SIGTERM)

        self._set_pid(None)

    def _get_pid(self):
        try:
            with open(self.pid_file, 'r') as pid_file:
                return int(pid_file.read().strip())
        except IOError:
            return

    def _set_pid(self, pid):
        if pid is None:
            try:
                os.unlink(self.pid_file)
            except OSError as e:
                error('Cannot remove pid file %s: %s', self.pid_file, e)
        else:
            with open(self.pid_file, 'w') as pid_file:
                pid_file.write(str(pid))
