# -*- coding: utf-8 -*-

"""
Daemons tasks
=============

This module defines functions to create tasks used to control daemons. Daemon
must be created as instances of ``sett.daemon.Daemon``,
``sett.daemon.DaemonGroup`` and ``sett.daemon.Daemons``. Unless daemon lib is
explicitely disabled, A ``Daemons`` instance is created when sett sets itself
up. Daemons are added to this daemons manager at initialization (through
on_init calls).

A virtual task daemon(xx) is added to the Paver task loaders and create tasks
to control the daemons as needed.

Example:

>>> from sett import on_init
>>> from sett.daemon import Daemon
>>> @on_init
... def register_daemons(env):
...     env.daemons.register(Daemon(['uwsgi', '--wsgi', 'myproject.wsgi']), name='front')
...     env.daemons.register(Daemon(['celery', 'myproject.celery']), name='backend')
...     env.daemons.register(Daemon(['mail_spool']), name='backend')

    $ paver daemon(mail_spool) start
    $ paver daemon(backend) status
    $ paver daemon(uwsgi) reload
"""

import functools
import sys

from paver.easy import consume_nargs, task, environment, error, info

from sett.task_loaders import RegexpTaskLoader
from sett.utils.dispatch import Dispatcher
from sett.utils import task_name as rename_task


def daemon_auto_task(daemon_name, task_name=None):
    @task
    @consume_nargs(1)
    @rename_task(daemon_name)
    def daemon_ctl(args, daemons):
        try:
            daemon_group = daemons[daemon_name]
        except KeyError:
            error('No daemon or group named %s', daemon_name)
            return
        dd = DaemonsDispatcher(daemon_group)
        dd(*args)

    return daemon_ctl


environment.task_finders.extend([
    RegexpTaskLoader(
        r'^daemon\/(\w+)$',
        daemon_auto_task,
    ),
    RegexpTaskLoader(
        r'^daemon\((\w+)\)$',
        daemon_auto_task,
    )
])


class DaemonDispatcher(Dispatcher):
    def __init__(self, daemon):
        self.daemon = daemon

    def as_task(self, name=None):
        name = name or self.daemon.name

        @task
        @consume_nargs(1)
        @rename_task
        def daemon_dispatcher_task(args):
            self(*args)
        return daemon_dispatcher_task

    def run(self):
        self.daemon.run()

    @Dispatcher.on('restart', 1)
    def start(self):
        self.daemon.start()

    @Dispatcher.on('restart', -1)
    def stop(self):
        self.daemon.stop()

    def reload(self):
        self.daemon.reload()

    def status(self):
        sys.stdout.write('{}\n'.format(self.daemon.status()))


class DaemonsDispatcher(DaemonDispatcher):
    def status(self):
        for daemon in self.daemon:
            sys.stdout.write('* {:40}: {}\n'.format(str(daemon), daemon.status()))


def ctl_task(fn):
    """
    Deprecated. Use daemons.register and daemon(xx) virtual tasks instead

    >>> from sett.daemons import Daemon, ctl_task
    >>> @ctl_task
    ... def uwsgi(self):
    ...     return Daemon(['uwsgi', '--wsgi', 'myproject.wsgi'])

    translates to:

    >>> from sett import on_init
    >>> @on_init
    ... def register_daemons(env):
    ...     env.daemons.register(Daemon(['uwsgi', '--wsgi', 'myproject.wsgi']))


    $ paver uswgi status

    translate to

    $ paver daemon(uwsgi) status
    """
    daemons = fn()
    if not daemons:
        return

    @task
    @consume_nargs(2)
    @functools.wraps(fn)
    def ctl(args):
        info('%s', ctl_task.__help__)
        arg, service = args
        dd = DaemonsDispatcher(daemons[service])
        dd(arg)
    return ctl


def daemon_task(fn):
    """
    Deprecated. Use daemons.register and daemon(xx) virtual tasks instead

    >>> from sett.daemons import Daemon, daemon_task
    >>> @daemon_task
    ... def ctl(self):
    ...     return Daemons(
    ...         Daemon(['uwsgi', '--wsgi', 'myproject.wsgi']),
    ...         Daemon(['celery', 'myproject.celery']),
    ...     )

    translates to:

    >>> from sett import on_init
    >>> @on_init
    ... def register_daemons(env):
    ...     env.daemons.register(Daemon(['uwsgi', '--wsgi', 'myproject.wsgi']))
    ...     env.daemons.register(Daemon(['celery', 'myproject.celery']))


    $ paver ctl uswgi status

    translate to

    $ paver daemon(uwsgi) status
    """

    daemon = fn()
    if not daemon:
        return

    @task
    @consume_nargs(1)
    @functools.wraps(fn)
    def daemon_ctl(args):
        info('%s', daemon_task.__help__)
        dd = DaemonDispatcher(daemon)
        dd(*args)

    return daemon_ctl
