#!/usr/bin/env python
# -*- coding: utf-8 -*-

from paver.easy import path, task, call_task, needs, info, debug

from sett import ROOT, defaults, which
from sett.utils import optional_import
from sett.deploy_context import DeployContext

jinja2 = optional_import('jinja2')

DeployContext.register(which=which)


@task
@needs(['wheel'])
def push(options):
    """Pushes the archive in the enix repo"""
    call_task('sdist')
    call_task('upload', options={
        'repository': defaults.PYPI_REPOSITORY,
    })


@task
def nginx_conf(paths):
    """
    Generates etc/nginx.conf
    """
    templates.render(defaults.NGINX_TEMPLATE,
                     paths.ETC.joinpath('nginx.conf'),
                     DeployContext({
                         'options': {
                             'FORCE_REWRITE': False,
                         }
                     }))


class JinjaTemplatesRenderer(object):
    def __init__(self):
        self._jinja_instance = None
        self._hooks = []

    def register_hook(self, hook):
        debug('Register template hook, %s', hook)
        self._hooks.append(hook)
        if self._jinja_instance is not None:
            hook(self._jinja_instance)
        return hook

    def _apply_hooks(self, jinja):
        for hook in self._hooks:
            hook(self._jinja_instance)

    @property
    def jinja_instance(self):
        if self._jinja_instance is None:
            debug('Creating jinja instance')
            location = path(__file__).dirname().joinpath('templates')
            debug('Loading builtin templates from %s', location)
            loader = jinja2.FileSystemLoader(location)  # Built in templates

            if ROOT.joinpath(defaults.DEPLOY_TEMPLATES_DIR).isdir():
                location = ROOT.joinpath(defaults.DEPLOY_TEMPLATES_DIR)
                debug('Loading templates from %s', location)
                # Load users or builtin
                loader = jinja2.ChoiceLoader([
                    jinja2.FileSystemLoader(location),
                    loader,
                    jinja2.PrefixLoader({
                        'super': loader,
                    })
                ])

            self._jinja_instance = jinja = jinja2.Environment(loader=loader)
            self._apply_hooks(jinja)

        return self._jinja_instance

    def get_template(self, template_name):
        return self.jinja_instance.get_template(template_name)

    def render(self, template, destination_path, context):
        """
        Render a jinja2 template into a file
        """
        destination_path = ROOT.joinpath(destination_path)
        destination_path.dirname().makedirs()

        template = self.get_template(template)
        rendered = template.render(context)
        info('Writing %s', destination_path)
        with open(destination_path, 'w') as destination:
            destination.write(rendered)


templates = JinjaTemplatesRenderer()


@templates.register_hook
def add_as_bool_filter(jinja):
    jinja.filters['as_bool'] = lambda x: x if x != 'false' else False


def render_template(*args, **kw):
    return templates.render(*args, **kw)
