#!/usr/bin/env python
# -*- coding: utf-8 -*-


import getpass
import os
import grp
import collections
import itertools

from sett import ROOT, defaults
from paver.easy import environment


class DeployContextInstance(collections.Mapping):
    def __init__(self, values):
        self.values = values

    def __repr__(self):
        return repr(self.values)

    def __getitem__(self, item):
        try:
            return self._get(item)
        except KeyError:
            raise KeyError('''Missing key ``{name}``: try defining it:
from sett import DeployContext
@DeployContext.register
def set_{name_fn}():
    return {{
        '{name}': None
    }}
'''.format(name=item, name_fn=item.replace('.', '_')))

    def _get(self, item):
        values = self.values
        for key in item.split('.'):
            values = values[key]
        return values

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def push(self, values):
        new_values = dict(self.values)
        new_values.update(values)
        return DeployContextInstance(new_values)


class DeployContextFactory(object):
    def __init__(self, **kw):
        self._providers = []
        self._defaults = [kw]

    def __iter__(self):
        return itertools.chain(self._defaults, self._providers)

    def register_default(self, provider_fn=None, **kw):
        provider = self._parse(provider_fn, kw)
        self._defaults.insert(0, provider)
        return provider

    def register(self, provider_fn=None, **kw):
        provider = self._parse(provider_fn, kw)
        self._providers.append(provider)
        return provider

    def _parse(sekf, provider_fn, kwargs):
        if provider_fn is None:
            assert kwargs
            return kwargs

        assert callable(provider_fn)
        assert not kwargs
        return provider_fn

    def __call__(self, values=None):
        context = {}
        if values:
            context.update(values)

        for provider in self:
            if callable(provider):
                provider = provider()

            assert isinstance(provider, collections.Mapping)

            for key, value in provider.items():
                if key not in context:
                    if type(value) in (list, dict):
                        # copy the content of list and dict in order to not erase the original
                        value = type(value)(value)
                    context[key] = value
                    continue

                current = context[key]
                if ((isinstance(value, (dict, list)) or isinstance(current, (dict, list))) and
                        type(current) != type(value)):
                    raise TypeError('Type mismatch: Expected a %r for %s but got a %r' %
                                    (type(current), key, type(value)))

                if isinstance(value, dict):
                    context[key].update(value)
                elif isinstance(value, list):
                    context[key].extend(value)
                else:
                    context[key] = value

        return DeployContextInstance(context)

DeployContext = DeployContextFactory()


@DeployContext.register_default
def default_context():
    setup_options = environment.get_task('setup_options')
    if not setup_options.called:
        return {}

    name = environment.options.setup.name.lower()
    return {
        'NAME': name,
        'domain': defaults.DOMAIN_TEMPLATE.format(name=name, env=environment)
    }


@DeployContext.register_default
def running_env():
    return {
        'uid': os.getuid(),
        'gid': os.getgid(),
        'ROOT': ROOT,
    }


@DeployContext.register_default
def running_resolved_env():
    try:
        uid = getpass.getuser()
    except KeyError:
        uid = os.getuid()

    try:
        gid = grp.getgrgid(os.getgid()).gr_name
    except KeyError:
        gid = os.getgid()

    return {
        'UID': uid,
        'GID': gid,
    }
