# -*- coding: utf-8 -*-

from sett import ROOT, DeployContext
from paver.easy import task, debug, consume_args, sh, info, consume_nargs, error

from sett.daemon import Daemon
from sett.daemon_tasks import DaemonDispatcher
from sett.paths import LOGS


PIDFILE = ROOT.joinpath('var/run/supervisor.pid')


@task
@consume_args
def supervisorctl(args, which, paths):
    """Wrapper to launch supervisorctl"""
    sh([
        which.supervisorctl,
        '-c', paths.ETC.joinpath('supervisord/supervisorctl.conf'),
    ] + args)


@task
@consume_nargs(1)
def supervisord(args, which, paths):
    """Controls the supervisord daemon"""
    supervisord = Daemon(
        [which.supervisord, '-c', paths.ETC.joinpath('supervisord/supervisord.conf')],
        daemonize=lambda pidfile: ['--pidfile', pidfile],
        pid_file=PIDFILE,
        reload='SIGHUP',
    )
    dd = DaemonDispatcher(supervisord)
    dd(*args)


@task
def supervisord_conf(daemons, paths):
    """
    The configuration files generator. It generates a list of files under
    etc/supervisord, which are all included in supervisord.conf. A file is
    written for each daemon and for each group. It also writes
    supervisorctl.conf which is used by supervisorctl and is not included in
    supervisord.conf

    etc/supervisord
    ├── daemons
    │   ├── nginx.conf
    │   ├── api.conf
    │   ├── celery-worker.conf
    │   └── web.conf
    ├── groups
    │   ├── backend.conf
    │   └── frontend.conf
    ├── supervisorctl.conf
    └── supervisord.conf
    """
    scw = SupervisordConfigWriter(paths.ETC.joinpath('supervisord'), daemons, DeployContext())
    scw()


class SupervisordConfigWriter(object):
    def __init__(self, root, daemons, context, socket=None):
        self.root = root
        self.daemons = daemons
        self.context = context
        self._supervisord_context = self.context.get('supervisord') or {}
        self._socket = None

    @property
    def socket(self):
        return self._socket or ROOT.joinpath('var/run/supervisor.sock')

    def __call__(self):
        if not self.daemons:
            raise ValueError('No daemons to write')

        conf_files = []
        conf_files.extend(self._write_group_confs())
        conf_files.extend(self._write_daemons_confs())

        self._write_daemon_conf(conf_files)
        self._write_client_conf()

    def _get_dir(self, name):
        directory = self.root.joinpath(name)
        directory.makedirs_p()
        return directory

    def _write_config(self, path, sections):
        debug('Writting %s', path)
        with open(path, 'w') as supervisord_conf_file:
            for header, instructions in sections.items():
                instructions.update(self._supervisord_context.get(header) or [])
                supervisord_conf_file.write('[{}]\n'.format(header))
                for instruction, value in instructions.items():
                    if isinstance(value, bool):
                        value = str(value).lower()
                    elif not value:
                        debug('Skipping empty %s', instruction)
                        continue

                    supervisord_conf_file.write('{}={}\n'.format(instruction, value))
                supervisord_conf_file.write('\n')

    def _write_client_conf(self):
        supervisorctl_conf = self.root.joinpath('supervisorctl.conf')
        self._write_config(supervisorctl_conf, {
            'supervisorctl': {
                'serverurl': 'unix://' + self.socket,
            },
        })

    def _write_daemon_conf(self, includes):
        supervisord_conf = self.root.joinpath('supervisord.conf')
        self._write_config(supervisord_conf, {
            'rpcinterface:supervisor': {
                'supervisor.rpcinterface_factory': 'supervisor.rpcinterface:make_main_rpcinterface',
            },
            'unix_http_server': {
                'file': self.socket,
                'chown': '{UID}:{GID}'.format(**self.context),
            },
            'supervisord': {
                'logfile': LOGS.joinpath('supervisord.log'),
                'pidfile': PIDFILE,
            },
            'include': {
                'files': ' '.join(includes),
            }
        })

    def _write_group_confs(self):
        groups_conf_dir = self._get_dir('groups')
        includes = []

        for group in self.daemons.groups():
            group_conf = groups_conf_dir.joinpath(group.name + '.conf')
            includes.append(group_conf)

            info('Writing group %s', group)
            self._write_config(group_conf, {
                'group:{}'.format(group.name): {
                    'programs': ','.join(d.name for d in group),
                }
            })
        return includes

    def _write_daemons_confs(self):
        daemons_conf_dir = self._get_dir('daemons')
        includes = []

        for daemon in self.daemons:
            daemon_conf = daemons_conf_dir.joinpath(daemon.name + '.conf')
            includes.append(daemon_conf)

            info('Writing daemon %s', daemon)
            self._write_config(daemon_conf, {
                'program:{}'.format(daemon.name): {
                    'command': daemon.command,
                    'environment': ','.join('{}="{}"'.format(key, val) for key, val in daemon.environ.items())
                },
            })
        return includes


def supervisord_task(fn):
    """
    Deprecated: use supervisord_conf instead and register your daemons
    """
    error('%s', supervisord_task.__doc__)
