# -*- coding: utf-8 -*-

import signal
from paver.easy import consume_args, error, sh, path, task, needs, info

from sett import ROOT, which, defaults
from sett.utils.dispatch import Dispatcher
from sett.deploy import templates


@task
@needs('setup_options')
@consume_args
def systemctl(args, options, daemons):
    try:
        daemon_group = daemons[args[0]]
    except KeyError:
        error('No daemon named %s', args[0])
        return

    commands = args[1:]
    for daemon in daemon_group:
        sdd = SystemdDispatcher(daemon.name)
        for command in commands:
            sdd(command)


class SystemdDispatcher(Dispatcher):
    def __init__(self, name, local_service_path=None, system_service_path=None):
        self.name = name
        self._local_service_path = local_service_path
        self._system_service_path = system_service_path

    @property
    def local_service_path(self):
        return self._local_service_path or ROOT.joinpath('etc/systemd.service')

    @property
    def system_service_path(self):
        return (self._system_service_path or
                path('/etc/systemd/system/').joinpath('{}.service'.format(self.name)))

    def _systemd(self, *args):
        sh([which.systemctl] + list(args))

    def _reload(self):
        self._systemd('daemon-reload')

    @Dispatcher.on('restart', 2)
    @Dispatcher.on('up', 2)
    def start(self):
        """Start the service"""
        self._systemd('start', self.name)

    @Dispatcher.on('restart', 1)
    def stop(self):
        """Stop the service"""
        self._systemd('stop', self.name)

    @Dispatcher.on('install', 1)
    @Dispatcher.on('up', 1)
    def link(self):
        """Creates a link in the system"""
        try:
            self.local_service_path.symlink(self.system_service_path)
        except OSError:
            error('Cannot write %s, sudo maybe?', self.system_service_path)
            info('# ln -s {target} {link}'.format(
                target=self.local_service_path,
                link=self.system_service_path,
            ))
        else:
            self._reload()

    @Dispatcher.on('uninstall', 1)
    def unlink(self):
        print('unlink')
        self._reload()

    @Dispatcher.on('install', 2)
    @Dispatcher.on('up', 3)
    def enable(self):
        self._systemd('enable', self.name)

    @Dispatcher.on('uninstall', 2)
    def disable(self):
        self._systemd('disable', self.name)


@templates.register_hook
def add_as_signal(jinja):
    def as_signal(value):
        if hasattr(signal, 'Signals') and isinstance(value, signal.Signals):
            return value.name
        return value
    jinja.filters['as_signal'] = as_signal


@task
def systemd_conf(daemons, deployment, paths, templates):
    """
    Generates etc/systemd.conf
    """
    systemd_conf_root = paths.ETC.joinpath('systemd/')
    deploy_context = deployment({
        'systemd': {},
    })
    systemd_context = deploy_context['systemd']

    for daemon in daemons:
        context = {'daemon': daemon}
        if daemon.name in systemd_context:
            context.update(systemd_context[daemon.name])

        templates.render(defaults.SYSTEMD_TEMPLATE,
                         systemd_conf_root.joinpath(daemon.name + '.service'),
                         deploy_context.push(context))
