#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
UWSGI Configration
==================

This module defines a configration generator for uwsgi. The configuration is
fetched from the DeployContext. It's written as ini or XML.

Uwsgi has builtin ini support. XML depends on libxml dev libs.

**Note**:

Don't use yaml in uWSGI, it's not yaml
"""


import collections
from xml.etree import ElementTree as ET

from paver.easy import task, info, debug, consume_nargs, call_task
from paver.deps.six import text_type, string_types

from sett import which, ROOT, defaults
from sett.daemon import Daemon
from sett.paths import LOGS
from sett.pip import VENV_DIR
from sett.deploy_context import DeployContext


CONFIG_ROOT = ROOT.joinpath('parts/etc/')


DeployContext.register_default(uwsgi={
    'master': True,
})


@DeployContext.register_default
def uwsgi_context_socket():
    if defaults.UWSGI_SOCKET_TYPE == 'unix':
        return {
            'uwsgi': {
                'socket': ROOT.joinpath('var', 'run', 'uwsgi.sock')
            }
        }
    else:
        return {
            'uwsgi': {
                'http': '{}:{}'.format(defaults.HTTP_WSGI_IP, defaults.HTTP_WSGI_PORT),
            }
        }


@task
def log_dir():
    """Ensure that the log directory exists"""
    if not LOGS.exists():
        LOGS.makedirs()


def Element(tag, text):
    el = ET.Element(tag)
    el.text = text_type(text)
    return el


@task
def uwsgi_conf():
    """
    Generates parts/uwsgi/uwsgi.xml
    """
    context = DeployContext({
        'uwsgi': {
            'processes': 1,
        },
        'env': [],
        'locations': {},
        'directories': {},
        'pythonpath': [
            ROOT,
        ],
    })

    module, name = context['wsgi_application'].rsplit('.', 1)
    config = {
        'module': '{}:{}'.format(module, name),
        'logto': LOGS.joinpath('uwsgi.log'),
        'home': VENV_DIR,
        'env': context['env'],
        'pythonpath': context['pythonpath'],
    }
    config.update(context['uwsgi'])

    if 'uwsgi.socket' in context:
        config.update({
            'chown-socket': '{UID}:{GID}'.format(**context),
            'chmod-socket': '660',
        })

    if defaults.STATIC_SERVER == 'uwsgi':
        config['static-map'] = [
            '{}={}'.format(context['locations'][key],
                           context['directories'][key])
            for key in context['locations']
        ]

    config.update(defaults.UWSGI_EXTRA)
    ouput_writer = get_uwsgi_output()
    ouput_writer.write(config)


def get_uwsgi_output():
    out = defaults.UWSGI_OUTPUT_FORMAT
    if out is None:
        out = 'ini'
        debug('Guessing %s output', out)

    if out == 'ini':
        return INIOutput(CONFIG_ROOT.joinpath('uwsgi.ini'))
    elif out == 'xml':
        return XMLOutput(CONFIG_ROOT.joinpath('uwsgi.xml'))

    raise NotImplementedError('No format named {}'.format(out))


class WSGIOutput(object):
    @classmethod
    def provider(cls, fn):
        def wrapper(destination):
            return cls(fn, destination)
        return wrapper

    def __init__(self, provider, destination):
        self.provider = provider
        self.destination = destination

    @property
    def config_path(self):
        assert isinstance(self.destination, string_types)
        return self.destination

    def write(self, config):
        if isinstance(self.destination, string_types):
            self.destination.dirname().makedirs()
            info('Writing %s', self.destination)
            with open(self.destination, 'w') as output:
                self.provider(config, output)
        else:
            self.provider(config, self.destination)


@WSGIOutput.provider
def INIOutput(config, output):
    output.write(u'[uwsgi]\n')
    items = list(config.items())
    items.sort()

    for k, v in items:
        if isinstance(v, bool):
            if not v:
                continue
            v = 'true'
        if isinstance(v, string_types) or not isinstance(v, collections.Iterable):
            v = [v]
        for value in v:
            output.write('{} = {}\n'.format(k, value))


@WSGIOutput.provider
def XMLOutput(config, output):
    root = ET.Element('uwsgi')
    for tag_name, text in config.items():
        if isinstance(text, list):
            for text_ in text:
                root.append(Element(tag_name, text_))
        elif isinstance(text, bool):
            root.append(ET.Element(tag_name))
        else:
            root.append(Element(tag_name, text))

    ET.ElementTree(root).write(output)


if 'uwsgi' in which:
    @task
    @consume_nargs(1)
    def daemon(daemons, which, args):
        """
        Deprecated. Use daemon(uwsgi) virtual task and explicitely declare the daemons.

        @on_init
        def add_uwsgi(env):
            from sett.uwsgi import get_uwsgi_output
            env.daemons.register(Daemon(
                [env.which.uwsgi, get_uwsgi_output().config_path],
                name='uwsgi',
                daemonize=lambda pidfile: ['--pidfile', pidfile, '--daemonize', '/dev/null'],
        ))
        """
        info('%s', daemon.__doc__)
        daemons.register(Daemon([which.uwsgi, get_uwsgi_output().config_path],
                                name='uwsgi',
                                daemonize=lambda pidfile: ['--pidfile', pidfile, '--daemonize', '/dev/null']))
        call_task('daemon(uwsgi)', args=args)
